/*******************************************************************************
 * Copyright 2013 Guillaume Saulnier-Comte
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/
package gsauln.decisiontree.factory;

import gsauln.decisiontree.DecisionTree;
import gsauln.decisiontree.Forest;
import gsauln.decisiontree.interfaces.IPoint;
import gsauln.decisiontree.interfaces.ITrainingPoint;
import gsauln.decisiontree.interfaces.ITrainingSetGenerator;
import gsauln.decisiontree.logic.Gardener;
import gsauln.decisiontree.state.interfaces.ITrainingState;

import java.io.Serializable;
import java.util.Collection;
import java.util.Random;

/**
 * Interface specifying the functions required by a Decision Tree Factory. 
 * @param <TP> Type of training points used during the creation of the trees.
 * @param <P> Type of point used with the decision tree during query.
 * @param <LC> Type of content present in the leafs of the tree.
 * @param <TS> Type of training state used during the creation of the trees.
 */
public interface IDecisionTreeFactory<TP extends ITrainingPoint<P, ?>, 
										P extends IPoint<?, ?>, 
										LC extends Serializable, 
										TS extends ITrainingState<TS, TP>> extends Serializable
{
	/**
	 * Return a decision tree created from the set of training points.
	 * @param set set of training points to create the decision tree from.
	 * @return a decision tree created from the set of training points.
	 */
	DecisionTree<P, LC> growTree(Collection<TP> set);

	/**
	 * Return a forest of containing numberOfTrees decision trees created from the set of points and initial trainingState.
	 * @param trainingSet training set object to get training data for each tree.
	 * @param numberOfTrees number of trees to create in the forest.
	 * @return a forest of containing numberOfTrees decision trees created from the set of points and initial trainingState.
	 */
	Forest<P, LC> growForest(final ITrainingSetGenerator<? extends Collection<TP>, TP, P> trainingSet, int numberOfTrees);
	
	/**
	 * Return the gardener used to create the trees and forests of this factory.
	 * @param random random number generator that should be used in the creation of the gardener to insure the 
	 * reproducibility of results.
	 * @return  the gardener used to create the trees and forests of this factory.
	 */
	Gardener<TP, P, LC, TS> getGardener(Random random);
	
	/**
	 * Return the initial training state used to create a tree.
	 * @param random random number generator that should be used in the creation of the training state to insure the 
	 * reproducibility of results.
	 * @return the initial training state used to create a tree.
	 */
	TS getInitialTrainingState(Random random);
	
}
