/*
 *   Copyright 2011 Guillaume Saulnier-Comte
 *
 *   Licensed under the Apache License, Version 2.0 (the "License");
 *   you may not use this file except in compliance with the License.
 *   You may obtain a copy of the License at
 *
 *       http://www.apache.org/licenses/LICENSE-2.0
 *
 *   Unless required by applicable law or agreed to in writing, software
 *   distributed under the License is distributed on an "AS IS" BASIS,
 *   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *   See the License for the specific language governing permissions and
 *   limitations under the License.
 */

package jparfor;


import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;

import jparfor.exceptions.MultiThreaderException;

/**
 * MultiThreader providing foreach functions enabling parallel computations of the content of for loops.  
 * The default number of threads is the number of processors.
 */
public final class MultiThreader
{

	private static int gNUMPROCS = getNumberOfAvailableProcessorsRuntimeCall();
	private static int gNUMTHREADS = gNUMPROCS;
	private static boolean gDEBUGMODE = false;

	/**
	 * Creates a new MultiThreader object and start the threadpool. 
	 */
	private MultiThreader() {}

	/**
	 * Execute the functor on all inputs.  All the inputs from the inputs iterable object are created and placed in
	 * a collection for multithread processing.  The results are returned in the same order that the inputs iterable
	 * created them.
	 * (If DebugMode is true, no multithreading is performed and a simple for loop is used).
	 * @param <I> Type of input to the functor.
	 * @param <R> Type of results from the functor.
	 * @param inputs Collection of inputs to send to the functor.
	 * @param functor Functor that will compute results from inputs.
	 * @return a Collection containing all the results calculated on the inputs with the functor.
	 */
	public static <I, R> ArrayList<R> foreach(Iterable<I> inputs, final Functor<I, R> functor)
	{
		if (gDEBUGMODE)
		{
			return foreachDebug(inputs, functor);
		}
		else
		{
			return foreachMT(inputs, -1, functor);
		}
	}

	/**
	 * Execute the functor on all inputs.  Only concurrent inputs from the inputs iterable object are created and placed in
	 * a collection for multithread processing at one time.  Multiple batch are created until all inputs from the iterable object have
	 * been computed.  The results are returned in the same order that the inputs iterable created them.  This function can be used if
	 * the inputs size for the functor is memory intensive and one cannot afford to create them all at the same time.  Note that
	 * the results must still be able to hold in memory all at the same time.
	 * (If DebugMode is true, no multithreading is performed and a simple for loop is used).
	 * @param <I> Type of input to the functor.
	 * @param <R> Type of results from the functor.
	 * @param inputs Collection of inputs to send to the functor.
	 * @param concurrentSize Maximum number of inputs to process in a batch (if set to -1, all inputs will be computed at once).  
	 * 	The inputs are put in batches according to the order given by the iterator.
	 * @param functor Functor that will compute results from inputs.
	 * @return a Collection containing all the results calculated on the inputs with the functor.
	 */
	public static <I, R> ArrayList<R> foreach(Iterable<I> inputs, int concurrentSize, final Functor<I, R> functor)
	{
		if (gDEBUGMODE)
		{
			return foreachDebug(inputs, functor);
		}
		else
		{
			return foreachMT(inputs, concurrentSize, functor);
		}
	}
	
	
	/**
	 * Execute the functor on all inputs.  All the inputs from the inputs iterable object are created and placed in
	 * a collection for multithread processing.  The results are returned in the same order that the inputs iterable
	 * created them.
	 * @param <I> Type of input to the functor.
	 * @param <R> Type of results from the functor.
	 * @param inputs Collection of inputs to send to the functor.
	 * @param functor Functor that will compute results from inputs.
	 * @return a Collection containing all the results calculated on the inputs with the functor.
	 */
	protected static <I, R> ArrayList<R> foreachMT(Iterable<I> inputs, final Functor<I, R> functor)
	{
		return foreachMT(inputs, -1, functor);
	}

	/**
	 * Execute the functor on all inputs.  Only concurrent inputs from the inputs iterable object are created and placed in
	 * a collection for multithread processing at one time.  Multiple batch are created until all inputs from the iterable object have
	 * been computed.  The results are returned in the same order that the inputs iterable created them.  This function can be used if
	 * the inputs size for the functor is memory intensive and one cannot afford to create them all at the same time.  Note that
	 * the results must still be able to hold in memory all at the same time.
	 * @param <I> Type of input to the functor.
	 * @param <R> Type of results from the functor.
	 * @param inputs Collection of inputs to send to the functor.
	 * @param concurrentSize Maximum number of inputs to process in a batch (if set to -1, all inputs will be computed at once).  
	 * 	The inputs are put in batches according to the order given by the iterator.
	 * @param functor Functor that will compute results from inputs.
	 * @return a Collection containing all the results calculated on the inputs with the functor.
	 */
	protected static <I, R> ArrayList<R> foreachMT(Iterable<I> inputs, int concurrentSize, final Functor<I, R> functor)
	{
		ArrayList<R> results = new ArrayList<R>();

		ArrayList<I> concurrentInputs = new ArrayList<I>();
		for (I input : inputs)
		{
			concurrentInputs.add(input);
			if (concurrentInputs.size() == concurrentSize)
			{
				results.addAll(executeTasks(concurrentInputs, functor));
				concurrentInputs = new ArrayList<I>();
			}
		}
		results.addAll(executeTasks(concurrentInputs, functor));

		return results;
	}
	
	/**
	 * Normal for loop to be used when the debug mode is true.
	 * @param <I> Type of input to the functor.
	 * @param <R> Type of results from the functor.
	 * @param inputs Collection of inputs to send to the functor.
	 * @param functor Functor that will compute results from inputs.
	 * @return a Collection containing all the results calculated on the inputs with the functor.
	 */
	protected static <I, R> ArrayList<R> foreachDebug(Iterable<I> inputs, final Functor<I, R> functor)
	{
		ArrayList<R> results = new ArrayList<R>();
		for (I input : inputs)
		{
			R result = functor.function(input);
			results.add(result);
		}
		return results;
	}

	/**
	 * Uses the threadpool to compute the function of the functor on all inputs. (isStarted() must be true).
	 * @param <I> Type of input to the functor.
	 * @param <R> Type of results from the functor.
	 * @param inputs Collection of inputs to send to the functor.
	 * @param functor Functor that will compute results from inputs.
	 * @return a Collection containing all the results calculated on the inputs with the functor.
	 */
	private static <I, R> ArrayList<R> executeTasks(ArrayList<I> inputs, final Functor<I, R> functor)
	{
		ExecutorService threadPool = Executors.newFixedThreadPool(gNUMTHREADS);
		
		ArrayList<R> results = null;
		
		// create the tasks
		ArrayList<Callable<IndexedResult<R>>> tasks = new ArrayList<Callable<IndexedResult<R>>>(inputs.size());
		for (int i = 0 ; i < inputs.size(); i++)
		{
			final int index = i;
			final I input = inputs.get(index);
			tasks.add(new Callable<IndexedResult<R>>()
			{
				@Override
				public IndexedResult<R> call()
				{
					R result = functor.function(input);
					IndexedResult<R> ir = new IndexedResult<R>(index, result);
					return ir;
				}
			});
		}
		
		// Do the computations
		try
		{
			List<Future<IndexedResult<R>>> futures = null;
			futures = threadPool.invokeAll(tasks);
			
			// get the results in order
			results = new ArrayList<R>(inputs.size());
			results.addAll(Collections.nCopies(inputs.size(), (R)null));
			for (Future<IndexedResult<R>> future : futures)
			{
				IndexedResult<R> ir = future.get();
				results.set(ir.getIndex(), ir.getResult());
			}
		}
		catch (InterruptedException e)
		{
			throw new MultiThreaderException(e.toString());
		}
		catch (ExecutionException e)
		{
			throw new MultiThreaderException(e.toString());
		}
		finally
		{
			threadPool.shutdown();
		}
		return results;
	}

	/**
	 * Return the number of processors used by the multithreader.
	 * @return the number of processors used by the multithreader.
	 */
	public static int getNumberOfAvailableProcessors()
	{
		return gNUMPROCS;
	}

	/**
	 * Return the number of processors available in the JVM.
	 * @return the number of processors available in the JVM.
	 */
	private static int getNumberOfAvailableProcessorsRuntimeCall()
	{
		Runtime runtime = Runtime.getRuntime();
		int numberOfProcessors = runtime.availableProcessors();
		return numberOfProcessors;
	}
	
	/**
	 * Return the number of threads that are created for parallel computations.
	 * @return the number of threads that are created for parallel computations.
	 */
	public static int getNumberOfThreads()
	{
		return gNUMTHREADS;
	}
	
	/**
	 * Set the number of threads to be created for parallel computations.
	 * @param threads number of threads to be created for parallel computations.
	 */
	public static void setNumberOfThreads(int threads)
	{
		gNUMTHREADS = threads;
	}
	
	/**
	 * Set the debug mode variable. (Default = false).
	 * @param debug if true, no parallel computations are performed and a simple for loop is used.
	 * This enables the trace for errors through the code making it easier to debug.  If false, parallel computations are used.
	 */
	public static void setDebugMode(boolean debug)
	{
		gDEBUGMODE = debug;
	}
	
	/**
	 * Return the status of the debug mode (true or false).
	 * @return the status of the debug mode (true or false).
	 */
	public static boolean getDebugMode()
	{
		return gDEBUGMODE;
	}

}
