/*
 *   Copyright 2011 Guillaume Saulnier-Comte
 *
 *   Licensed under the Apache License, Version 2.0 (the "License");
 *   you may not use this file except in compliance with the License.
 *   You may obtain a copy of the License at
 *
 *       http://www.apache.org/licenses/LICENSE-2.0
 *
 *   Unless required by applicable law or agreed to in writing, software
 *   distributed under the License is distributed on an "AS IS" BASIS,
 *   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *   See the License for the specific language governing permissions and
 *   limitations under the License.
 */

package jparfor;


import java.util.Iterator;

import jparfor.exceptions.RangeException;



/**
 * Implement an iterable range that is computed lazily.
 * 
 * The range will start at the start value (inclusive) and end at the end value (exclusive) doing increments
 * of step size each time towards the end.
 * 
 * A for loop as
 * for (int i = 5; i < 22; i += 3)
 * 
 * can now be replaced by
 * for (int i : new Range(5, 22, 3))
 * 
 */
public class Range implements Iterable<Integer>
{

	private int fStart;
	private int fEnd;
	private int fStep;
	private int fSide;
	
	/**
	 * Create an iterable range object with the following properties.
	 * @param start start value (inclusive).
	 * @param end end value (exclusive).
	 * @param step size of the increment between values (> 0).
	 */
	public Range(int start, int end, int step)
	{
		if (step <= 0)
		{
			throw new RangeException("Step size should be positive! step = "+step);
		}
		init(start, end, step);
	}
	
	/**
	 * Create an iterable range object with the following properties (step = 1).
	 * @param start start value (inclusive).
	 * @param end end value (exclusive).
	 */
	public Range(int start, int end)
	{
		init(start, end, 1);
	}
	
	/**
	 * Create an iterable range object with the following properties (start = 0, step = 1).
	 * @param end end value (exclusive).
	 */
	public Range(int end)
	{
		init(0, end, 1);
	}
	
	/**
	 * Create an iterable range object with the following properties [start, end) with step increments.
	 * @param start start value (inclusive).
	 * @param end end value (exclusive).
	 * @param step increment between values.
	 */
	private void init(int start, int end, int step)
	{
		fStart = start;
		fEnd = end;
		fStep = step;
		if (fStart <= fEnd)
		{
			fSide = 1;
		}
		else
		{
			fSide = -1;
		}
	}
	
	/**
	 * Return the start of the range object.
	 * @return the start of the range object.
	 */
	public int getStart()
	{
		return fStart;
	}
	
	/**
	 * Return the end of the range object.
	 * @return the end of the range object.
	 */
	public int getEnd()
	{
		return fEnd;
	}
	
	/**
	 * Return the step of the rage object.
	 * @return the step of the rage object.
	 */
	public int getStep()
	{
		return fStep;
	}
	
	@Override
	public Iterator<Integer> iterator()
	{
		return new RangeIterator();
	}
	
	/**
	 * Iterator object for the range object.
	 */
	public class RangeIterator implements Iterator<Integer>
	{

		private int fCurrent;
		
		/**
		 * Creates a new iterator for the range object.
		 */
		public RangeIterator()
		{
			fCurrent = fStart;
		}
		
		@Override
		public boolean hasNext()
		{
			return fSide * fCurrent < fSide * fEnd;
		}

		@Override
		public Integer next()
		{
			if (fSide * fCurrent >= fSide * fEnd)
			{
				throw new RangeException("Out of Range!");
			}
			int index = fCurrent;
			fCurrent += fSide * fStep;
			return index;
		}

		@Override
		public void remove()
		{
			throw new RangeException("Not implemented!");
		}
		
	}

}
